#include <stdio.h>
#include <stdlib.h>
#include <ldap.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>
#include "radiusd.h"
#include "authldap.h"
#include "localip.h"

extern struct LDAP_s *ldaps;
extern int auth_port;


/* BH 2001040801
 * cycle ldapserver connection list
 */
void rr_ldaps(int qtype)
{
	switch(qtype)
	{
		case ALDAP_QTYPE_USER :
			ldaps -> user_s = ldaps -> user_s -> next;
			break;

		case ALDAP_QTYPE_CLIENT :
			ldaps -> client_s = ldaps -> client_s -> next;
			break;

		default :
			log(L_ERR, " rr_ldaps: unknown querytype");
			return;
	}
}


void ldap_unbind_ldaps(LDAPCON *ldc)
{
	int result;

	if (ldc -> ld != NULL)
	{
		result = ldap_unbind(ldc -> ld);
#ifdef ALDAP_DEBUG
			if (result == -1)
				log(L_LDAPB, "error unbinding!!");
#endif
		ldc -> ld = NULL;
	}
}


int ldap_bind_ldaps_nrr(LDAP **ld, int qtype)
{
	int		result = ALDAP_ERR_UNKNOWN,
				msgid;

	struct timeval
				timeout;

	LDAP		*ld1;
	LDAPMessage
				*msg;

	LDAPCON	*ldc;

	*ld = NULL;

	switch(qtype)
	{
		case ALDAP_QTYPE_USER :
			ldc = ldaps -> user_s;
			break;

		case ALDAP_QTYPE_CLIENT :
			ldc = ldaps -> client_s;
			break;

		default :
			log(L_ERR, " ldap_bind_ldaps_nrr: unknown querytype");
			return(ALDAP_ERR_QTYPE);
	}

	if (ldc -> ld == NULL)
	{
		DEBUG2(" ldap_bind_ldaps_nrr: server == %s, binddn == %s", ldc -> server, ldc -> binddn);
		if ((ld1 = ldap_init(ldc -> server, ALDAP_PORT)) != NULL)
		{
			if ((msgid = ldap_simple_bind(ld1, ldc -> binddn, ldc -> bindpasswd)) != -1)
			{
				timeout.tv_sec = ALDAP_MAXTIMEOUT;
				timeout.tv_usec = 0;
				switch (ldap_result(ld1, msgid, 0, &timeout, &msg))
				{
					case 0 :
						result = ALDAP_ERR_TIMEOUT;
						break;

					case -1 :
						result = ALDAP_ERR_RESULT;
						break;

					case LDAP_RES_BIND :

						/* check bind result and free msg-structure */
						switch (ldap_result2error(ld1, msg, 1))
						{
							case LDAP_INVALID_CREDENTIALS :
								result = ALDAP_ERR_CRED;
								break;

							case LDAP_SUCCESS :
								ldc -> ld = ld1;
								*ld = ld1;
								result = ALDAP_ERR_NOERR;
								break;

							default :
								result = ALDAP_ERR_UNKNOWN;
						}
						break;

					/* if resulttype was not 'bind',
					 * free msg-structure and return errorcode
					 */
					default :
						ldap_msgfree(msg);
						result = ALDAP_ERR_RESULTDIFF;
				}
			}
			else
				result = ALDAP_ERR_BIND;

			/* if something wrong happens, unbind */
			if (result != ALDAP_ERR_NOERR)
			{
				DEBUG2(" ldap_bind_ldaps_nrr: unbinding");
				ldap_unbind(ld1);
				result = ALDAP_ERR_BIND;
			}
		}
		else
			result = ALDAP_ERR_INIT;

	} /* if (ldc -> ld == NULL) */
	else
	{
		*ld = ldc -> ld;
		result = ALDAP_ERR_NOERR;
	}

	DEBUG2(" ldap_bind_ldaps_nrr: con %d, ld $%08x", ldc -> con, ldc -> ld);
	if (result != ALDAP_ERR_NOERR)
		DEBUG(" ldap_bind_ldaps_nrr: result == %d", result);

/*	rr_ldaps(qtype);*/
	return(result);
}

int ldap_bind_ldaps(LDAP **ld, int qtype)
{
	int result;

	result = ldap_bind_ldaps_nrr(ld, qtype);
	rr_ldaps(qtype);
	return(result);
}


/* last memory audit: BH2001072301 */
int ldap_compare_attr(LDAP *ld, char *dn, char *attr, char *val)
{
	int	msgid,
			result = ALDAP_ERR_UNKNOWN;

	LDAPMessage
			*msg;

	struct timeval
			timeout;

	if ((msgid = ldap_compare(ld, dn, attr, val)) != -1)
	{
		DEBUG2(" ldap_compare_attr: msgid == %d, dn == %s", msgid, dn);
		timeout.tv_sec = ALDAP_MAXTIMEOUT;
		timeout.tv_usec = 0;
		switch (ldap_result(ld, msgid, 0, &timeout, &msg))
		{
			case 0 :
				result = ALDAP_ERR_TIMEOUT;
				break;

			case -1 :
				result = ALDAP_ERR_RESULT;
				break;

			case LDAP_RES_COMPARE :
				DEBUG(" ldap_compare_attr: ldapresult == LDAP_RES_COMPARE");
				/* check compare result and free msg-structure */
				switch (msgid = ldap_result2error(ld, msg, 1))
				{
					case LDAP_COMPARE_TRUE :
						result = ALDAP_ERR_NOERR;
						break;

					case LDAP_COMPARE_FALSE :
						result = ALDAP_ERR_CMPFALSE;
						break;

					case LDAP_NO_SUCH_OBJECT :
						result = ALDAP_ERR_NOENTRY;
						break;

					default :
						DEBUG2(" ldap_compare_attr: ldaperror == %s", ldap_err2string(msgid));
						result = ALDAP_ERR_UNKNOWN;
				}
				break;

			/* if resulttype was not 'compare',
			 * free msg-structure and return errorcode
			 */
			default :
				ldap_msgfree(msg);
				result = ALDAP_ERR_RESULTDIFF;
		}
	}
	else
		result = ALDAP_ERR_CMP;

	if (result != ALDAP_ERR_NOERR)
		DEBUG(" ldap_compare_attr: result == %d", result);

	return(result);
}


/* last memory audit: BH2001072301 */
char *ldap_getstring(LDAP *ld, char *dn, char *attr, int *resret)
{
	struct timeval
			timeout;

	LDAPMessage
			*msg = NULL;

	int	res,
			result = ALDAP_ERR_UNKNOWN,
			msgid;
	
	char	**vals,
			*secret = NULL;

	if ((msgid = ldap_search(ld, dn, LDAP_SCOPE_BASE, NULL, NULL, 0)) != -1)
	{
		DEBUG2(" ldap_getstring: msgid == %d", msgid);
		timeout.tv_sec = ALDAP_MAXTIMEOUT;
		timeout.tv_usec = 0;
		switch (ldap_result(ld, msgid, 1, &timeout, &msg))
		{
			case 0 :
				result = ALDAP_ERR_TIMEOUT;
				break;

			case -1 :
				result = ALDAP_ERR_RESULT;
				break;

			case LDAP_RES_SEARCH_RESULT :
				DEBUG(" ldap_getstring: ldapresult == %s", "LDAP_RES_SEARCH_RESULT");
				/* check compare result and free msg-structure */
				switch (res = ldap_result2error(ld, msg, 0))
				{
					case LDAP_SUCCESS :
						DEBUG2(" ldap_getstring: ldapresult == %s", ldap_err2string(res));
						if ((vals = ldap_get_values(ld, msg, attr)) != NULL)
						{
							secret = strdup(*vals);
							ldap_value_free(vals);
							result = ALDAP_ERR_NOERR;
						}
						else
							result = ALDAP_ERR_NOATTR;
						break;

					case LDAP_NO_SUCH_OBJECT :
						DEBUG2(" ldap_getstring: ldapresult == %s", ldap_err2string(res));
						result = ALDAP_ERR_NOENTRY;
						break;

					default :
						DEBUG2(" ldap_getstring: ldapresult == %s", ldap_err2string(res));
				} /* switch (ldap_result2error( ... */
				ldap_msgfree(msg);
				break;

			default :
				ldap_msgfree(msg);
				result = ALDAP_ERR_RESULTDIFF;
		} /* switch (ldap_result( ... */
	}
	else
		result = ALDAP_ERR_SEARCH;

	if (resret != NULL)
		*resret = result;

	if (result != ALDAP_ERR_NOERR)
	{
		DEBUG2(" ldap_getstring: result == %d", result);
		return(NULL);
	}

	return(secret);
}


/*
 *	check pass, name and realm against
 *	LDAP directory
 *	BH 2001030101
 */
/* last memory audit: BH2001072301 */
int ldap_pass(char *name, char *passwd, UINT4 ipaddr, int mode)
{
	int	result = ALDAP_ERR_UNKNOWN,
			dnlen;

	LDAP	*ld;

	char	*dn,
			*realm,
			*prealm,
			*rpasswd;


	DEBUG2(" ldap_pass: name == \"%s\"", name);
	if ((passwd != NULL) && (strlen(passwd) > 0))
	{
		/* check if name contains realm
		 * FIXME chained realms not possible
		 */
		if ((prealm = strrchr(name, '@')) != NULL)
		{
			/* strip of realm from username and copy to separate string */
			*prealm++ = 0;
			if ((realm = (char*) malloc(strlen(prealm) + 1)) != NULL)
				strcpy(realm, prealm);
			else result = ALDAP_ERR_NOMEM;
		}
		else
		{
			/* user default realm from authldap.h */
			if ((realm = (char*) malloc(strlen(ALDAP_REALM) + 1)) != NULL)
				strcpy(realm, ALDAP_REALM);
			else result = ALDAP_ERR_NOMEM;
		}

		if (result == ALDAP_ERR_NOMEM)
		{
			DEBUG2(" ldap_pass: result == %d", result);
			return(result);
		}

		if ((result = ldap_client_realm(ipaddr, realm)) != ALDAP_ERR_NOERR)
		{
			log(L_ERR, " ldap_pass: @%s not %s", realm, client_name(ipaddr));
			switch (result)
			{
				case ALDAP_ERR_CMPFALSE :
					result = ALDAP_ERR_NOMEMBER;
					break;
				case ALDAP_ERR_NOENTRY :
					result = ALDAP_ERR_NOCLIENT;
					break;
			}
			free(realm);
			DEBUG2(" ldap_pass: result == %d", result);
			return(result);
		}
		result = ALDAP_ERR_UNKNOWN;

		dnlen = strlen(name) + strlen(realm) + strlen(ALDAP_DN) + 1;
		if (((dn = (char*) malloc(dnlen)) != NULL))
		{
			sprintf(dn, ALDAP_DN, name, realm);
			DEBUG(" ldap_pass: dn == %s", dn);

			if ((result = ldap_bind_ldaps(&ld, ALDAP_QTYPE_USER)) == ALDAP_ERR_NOERR)
			{
				DEBUG2(" ldap_pass: bind success");
				switch (mode)
				{
					case ALDAP_PASS_COMPARE :
						result = ldap_compare_attr(ld, dn, ALDAP_ATTR_UPASS, passwd);
						break;

					case ALDAP_PASS_READ :
						if ((rpasswd = ldap_getstring(ld, dn, ALDAP_ATTR_UPASS, &result)) != NULL)
						{
							strcpy(passwd, rpasswd);
							free(rpasswd);
						}
						break;

					default :
						result = ALDAP_ERR_PMODE;
				}

				/* BH2001050801
				 * unbind if connection got lost
				 * FIXME, that should be done at every ldapcall
				 */
				if (result == ALDAP_ERR_RESULT)
					ldap_unbind_ldaps(ldaps -> user_s);
			}
/*			else result = ALDAP_ERR_BIND;*/
			free(dn);
		}
		else result = ALDAP_ERR_NOMEM;
		free(realm);
	}
	else result = ALDAP_ERR_NOPASS;

	if (result != ALDAP_ERR_NOERR)
		DEBUG2(" ldap_pass: result == %d", result);

	return(result);
}


void fill_realm(REALM *rl, char *realm, int strip, char *iphost, int ipport)
{
	memset(rl, 0, sizeof(REALM));

	if (strcmp(iphost, "LOCAL") != 0)
		rl -> ipaddr = get_ipaddr(iphost);

	strNcpy(rl -> realm, realm, sizeof(rl -> realm));
	strNcpy(rl -> server, iphost, sizeof(rl -> server));

	rl -> auth_port = ipport;
	rl -> acct_port = ipport + 1;
	rl -> striprealm = strip;
	rl -> next = NULL;
}

/* last memory audit: BH2001071301 */
REALM *ldap_realm(char *realm)
{
	static REALM
			rl,
			*rlr;

	LDAP	*ld;

	LDAPMessage
			*msg;

	char	*dn,
			*iphost = NULL,
			**vals;

	int	dnlen,
			strip = 0,
			msgid,
			ipport = ALDAP_DEFPORT,
			result = ALDAP_ERR_NOERR;

	struct timeval
			timeout;


	if ((realm != NULL) && (strlen(realm) > 0))
	{
		if ((result = ldap_bind_ldaps(&ld, ALDAP_QTYPE_CLIENT)) == ALDAP_ERR_NOERR)
		{
			dnlen = strlen(realm) + strlen(ALDAP_LOCALHOST) + strlen(ALDAP_RDN) + 1;
			if ((dn = (char*) malloc(dnlen)) != NULL)
			{
				sprintf(dn, ALDAP_RDN, realm, ALDAP_LOCALHOST);
				DEBUG(" ldap_realm: dn == %s", dn);

				if ((msgid = ldap_search(ld, dn, LDAP_SCOPE_BASE, NULL, NULL, 0)) != -1)
				{
					DEBUG2(" ldap_realm: msgid == %d", msgid);
					msg = NULL;
					timeout.tv_sec = ALDAP_MAXTIMEOUT;
					timeout.tv_usec = 0;
					switch (ldap_result(ld, msgid, 1, &timeout, &msg))
					{
						case 0 :
							result = ALDAP_ERR_TIMEOUT;
							break;

						case -1 :
							result = ALDAP_ERR_RESULT;
							break;

						case LDAP_RES_SEARCH_RESULT :
							DEBUG2(" ldap_realm: ldapresult == LDAP_RES_SEARCH_RESULT");
							/* check search result and free msg-structure */
							switch (msgid = ldap_result2error(ld, msg, 0))
							{
								case LDAP_SUCCESS :
									/* read rewriting info from directory */
									if ((vals = ldap_get_values(ld, msg, ALDAP_REALM_RW)) != NULL)
									{
										DEBUG(" ldap_realm: rewriting %s => %s", realm, strchr(*vals, '=') + 1);
										/* if exists, recursivly call ldap_realm */
										rlr = ldap_realm(strchr(*vals, '=') + 1);
										/* free all memories */
										ldap_value_free(vals);
										ldap_msgfree(msg);
										free(dn);
										/* return structure */
										if (rlr != NULL)
											return(rlr);
										else
											return(NULL);
									}
									/* read ipserviceport from directory */
									if ((vals = ldap_get_values(ld, msg, ALDAP_IPPORT)) != NULL)
									{
										if ((ipport = atoi(*vals)) == 0) ipport = ALDAP_DEFPORT;
										ldap_value_free(vals);
									}
									/* read iphost from directory */
									if ((vals = ldap_get_values(ld, msg, ALDAP_IPHOST)) != NULL)
									{
										if ((iphost = strdup(*vals)) != NULL)
										{
											/* we possibly should strip off leading and trailing spaces here */
											if ((strcmp(ALDAP_LOCALHOST, iphost) == 0) && (ipport == auth_port))
											{
												free(iphost);
												iphost = NULL;
											}
										}
										ldap_value_free(vals);
									}
									/* if no host in directory or host eq localhost, use localhost ('LOCAL')*/
									if (iphost == NULL)
										iphost = strdup("LOCAL");
									/* strip or nostrip? from directory */
									if ((vals = ldap_get_values(ld, msg, ALDAP_STRIP)) != NULL)
									{
										if (strcmp(*vals, "strip") == 0) strip = 1;
										ldap_value_free(vals);
									}

									/* get and fill realm structure */
									if (iphost != NULL)
									{
										fill_realm(&rl, realm, strip, iphost, ipport);
										result = ALDAP_ERR_NOERR;
										DEBUG(" ldap_realm: %s, %s, 0x%08lx:%d, %d", rl.realm, rl.server, rl.ipaddr, rl.auth_port, rl.striprealm);
										free(iphost);
									}
									else
										result = ALDAP_ERR_NOMEM;
									break;

								case LDAP_NO_SUCH_OBJECT :
									fill_realm(&rl, realm, 0, "LOCAL", auth_port);
									result = ALDAP_ERR_NOERR;
									DEBUG(" ldap_realm: no proxying specified in LDAP, using local host for realm %s", realm);
									break;

								default :
									result = ALDAP_ERR_UNKNOWN;
							}
							ldap_msgfree(msg);
							break;

						/* if resulttype was not 'search',
						 * free msg-structure and return errorcode
						 */
						default :
							ldap_msgfree(msg);
							result = ALDAP_ERR_RESULTDIFF;
					}
				}
				else
					result = ALDAP_ERR_SEARCH;
				free (dn);
			}
			else
				result = ALDAP_ERR_NOMEM;
		}
	}
	else
		result = ALDAP_ERR_NOREALM;

	if (result != ALDAP_ERR_NOERR)
	{
		DEBUG(" ldap_realm: result == %d", result);
		return(NULL);
	}

	return &rl;
}

/* last memory audit: BH2001071301 */
CLIENT *ldap_client(UINT4 ipaddr)
{
	static CLIENT
			cl;

	LDAP	*ld;

	char	*dn,
			*secret = NULL,
			hstname[128];

	int	dnlen,
			result = ALDAP_ERR_UNKNOWN;

	if ((result = ldap_bind_ldaps(&ld, ALDAP_QTYPE_CLIENT)) == ALDAP_ERR_NOERR)
	{
		ipaddr2str(hstname, ipaddr);
		dnlen = strlen(hstname) + strlen(ALDAP_CDN) + 1;
		if ((dn = (char*) malloc(dnlen)) != NULL)
		{
			sprintf(dn, ALDAP_CDN, hstname);
			DEBUG(" ldap_client: dn == %s", dn);

			/* get and fill client structure */
			if ((secret = ldap_getstring(ld, dn, ALDAP_SECRET, NULL)) != NULL)
			{
				DEBUG2(" ldap_client: secret found");
				memset(&cl, 0, sizeof(CLIENT));
				cl.ipaddr = ipaddr;
				strNcpy(cl.secret, secret, sizeof(cl.secret));
				strNcpy(cl.longname, hstname, sizeof(cl.longname));
				strNcpy(cl.shortname, hstname, sizeof(cl.shortname));
				cl.next = NULL;
				result = ALDAP_ERR_NOERR;
				free(secret);
			}
			else
				result = ALDAP_ERR_NOENTRY;
			free (dn);
		}
		else
			result = ALDAP_ERR_NOMEM;
	}

	if (result != ALDAP_ERR_NOERR)
	{
		DEBUG(" ldap_client: result == %d", result);
		return(NULL);
	}

	return &cl;
}


/* BH 2001041701
 *	check if realm is allowed from specified client
 */
/* last memory audit: BH2001072301 */ 
int ldap_client_realm(UINT4 ipaddr, char *realm)
{
	int	result = ALDAP_ERR_UNKNOWN,
			dnlen;

	LDAP	*ld;

	char	*dn,
			hstname[128];

	ipaddr2str(hstname, ipaddr);
	dnlen = strlen(hstname) + strlen(ALDAP_CDN) + 1;
	if (((dn = (char*) malloc(dnlen)) != NULL))
	{
		sprintf(dn, ALDAP_CDN, hstname);
		if ((result = ldap_bind_ldaps(&ld, ALDAP_QTYPE_CLIENT)) == ALDAP_ERR_NOERR)
		{
			result = ldap_compare_attr(ld, dn, ALDAP_ATTR_REALM, realm);
		}
		else
			result = ALDAP_ERR_BIND;
		free(dn);
	}
	else
		result = ALDAP_ERR_NOMEM;

	if (result != ALDAP_ERR_NOERR)
		DEBUG(" ldap_client_realm: result == %d", result);

	return(result);
}
