#ifdef AUTHLDAP

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <lber.h>
#include "radiusd.h"

extern struct LDAP_s *ldaps;


TLDAP_RADIUS *cmp_radius_ldap_attr(char *ldattr, TLDAP_RADIUS *match_list)
{
	TLDAP_RADIUS *ml;

	for (ml = match_list; ml; ml = ml -> next)
		if (!strcmp(match_list -> attr, ldattr))
			return(match_list);
	return(NULL);
}



int ldap_user_find(char *name, char *checklist, char *replylist)
{
	int	result = ALDAP_ERR_UNKNOWN,
			dnlen,
			msgid;

	struct timeval
				timeout;

	LDAP	*ld;
	LDAPMessage
			*msg,
			*entry;
	BerElement
			*ber;

	char	*dn,
			*realm,
			*prealm,
			*attrname,
			**attrvalue;

	TLDAP_RADIUS
			*item;


	DEBUG2(" ldap_user_find: name == %s", name);
	/* check if name contains realm
	 * FIXME chained realms not possible
	 */
	if ((prealm = strrchr(name, '@')) != NULL)
	{
		/* strip of realm from username and copy to separate string */
		*prealm++ = 0;
		if ((realm = (char*) malloc(strlen(prealm) + 1)) != NULL)
			strcpy(realm, prealm);
		else result = ALDAP_ERR_NOMEM;
	}
	else
	{
		/* user default realm from authldap.h */
		if ((realm = (char*) malloc(strlen(ALDAP_REALM) + 1)) != NULL)
			strcpy(realm, ALDAP_REALM);
		else result = ALDAP_ERR_NOMEM;
	}

	if (result == ALDAP_ERR_NOMEM)
		return(result);

/* BH2001071101
 * check if realm on this client allowed to login
 */
/*
	if ((result = ldap_client_realm(ipaddr, realm)) != ALDAP_ERR_NOERR)
	{
		log(L_ERR, "request with realm %s from client %s not allowed or client unknown", realm, client_name(ipaddr));
		switch (result)
		{
			case ALDAP_ERR_CMPFALSE :
				result = ALDAP_ERR_NOMEMBER;
				break;
			case ALDAP_ERR_NOENTRY :
				result = ALDAP_ERR_NOCLIENT;
				break;
		}
		return(result);
	}
	result = ALDAP_ERR_UNKNOWN;
*/

	dnlen = strlen(name) + strlen(realm) + strlen(ALDAP_DN) + 1;
	if (((dn = (char*) malloc(dnlen)) != NULL))
	{
		sprintf(dn, ALDAP_DN, name, realm);
		DEBUG2(" ldap_user_find: dn == %s", dn);
		if ((result = ldap_bind_ldaps(&ld, ALDAP_QTYPE_USER)) == ALDAP_ERR_NOERR)
		{
			DEBUG2(" ldap_user_find: bind success");

			if ((msgid = ldap_search(ld, dn, LDAP_SCOPE_BASE, NULL, NULL, 0)) != -1)
			{
				msg = NULL;
				timeout.tv_sec = ALDAP_MAXTIMEOUT;
				timeout.tv_usec = 0;
				switch (ldap_result(ld, msgid, 1, &timeout, &msg))
				{
					case 0 :
						result = ALDAP_ERR_TIMEOUT;
						break;

					case -1 :
						result = ALDAP_ERR_RESULT;
						break;

					case LDAP_RES_SEARCH_RESULT :
						DEBUG2(" ldap_user_find: LDAP_RES_SEARCH_RESULT");
						/* check search result and free msg-structure */
						switch (msgid = ldap_result2error(ld, msg, 0))
						{
							case LDAP_SUCCESS :
								if ((entry = ldap_first_entry(ld, msg)) != NULL)
								{
									if ((attrname = ldap_first_attribute(ld, entry, &ber)) != NULL)
									{
										*checklist = 0;
										*replylist = 0;
										while (attrname != NULL)
										{
											attrvalue = ldap_get_values(ld, entry, attrname);
											if ((item = cmp_radius_ldap_attr(*attrvalue, ldaps -> check_item_map)) != NULL)
												sprintf(checklist + strlen(checklist), "%s = %s,", item -> radius_attr, *attrvalue);
											else if ((item = cmp_radius_ldap_attr(*attrvalue, ldaps -> reply_item_map)) != NULL)
												sprintf(replylist + strlen(replylist), "%s = %s,", item -> radius_attr, *attrvalue);
											ldap_value_free(attrvalue);
											attrname = ldap_next_attribute(ld, entry, ber);
										}
										if (strlen(checklist) > 0)
											*(checklist + strlen(checklist) - 1) = 0;
										if (strlen(replylist) > 0)
											*(replylist + strlen(replylist) - 1) = 0;
										ber_free(ber, 0);
										result = ALDAP_ERR_NOERR;
									}
									else
										result = ALDAP_ERR_FIRSTATTR;
								}
								else
									result = ALDAP_ERR_FIRSTENTRY;
								break;

							default :
								DEBUG2(" ldap_user_find: %s", ldap_err2string(msgid));
								result = ALDAP_ERR_UNKNOWN;
						}
						ldap_msgfree(msg);
						break;

					/* if resulttype was not 'search',
					 * free msg-structure and return errorcode
					 */
					default :
						ldap_msgfree(msg);
						result = ALDAP_ERR_RESULTDIFF;
				}
			}
			else
				result = ALDAP_ERR_SEARCH;

			/* BH2001050801
			 * unbind if connection got lost
			 * FIXME, that should be done at every ldapcall
			 */
			if (result == ALDAP_ERR_RESULT)
				ldap_unbind_ldaps(ldaps -> user_s);
		}
		free(dn);
	}
	else result = ALDAP_ERR_NOMEM;
	free(realm);

	if (result != ALDAP_ERR_NOERR)
		DEBUG(" !ldap_user_find: result == %d", result);

	return(result);
}

#endif